

function meas_sfcf()

    if params["sfcf"]["meas"]

        println(log_file,"Measuring Sfcf...")
        flush(log_file)

        @timeit "Sfcf" compute_propagators()
    end

    return nothing
end


"""
    function compute_propagators()

Computes the propagators for the first two components of spin.
The propagators are stored in Acs (t=0 to bulk), Bcs(bulk to t=T) and Ccs(t=0 to t=T)
"""
function compute_propagators()

    for f in 1:length(dpar)

        println(log_file,"Computing propagators from t=0 to the bulk for fermion "*fernames[f])

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 1, 1)
        println(log_file,"CG converged for c=1 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A11 .= Array(psi)
        global C11 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 2, 1)
        println(log_file,"CG converged for c=2 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A21 .= Array(psi)
        global C21 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 3, 1)
        println(log_file,"CG converged for c=3 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A31 .= Array(psi)
        global C31 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 1, 2)
        println(log_file,"CG converged for c=1 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A12 .= Array(psi)
        global C12 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 2, 2)
        println(log_file,"CG converged for c=2 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A22 .= Array(psi)
        global C22 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        niter = bndpropagator!(psi, U, dpar[f], dws, lp, params["Fermion"*fernames[f]]["maxiter"], params["Fermion"*fernames[f]]["tolerance"], 3, 2)
        println(log_file,"CG converged for c=3 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm2(x), +, dws.sr))
        global A32 .= Array(psi)
        global C32 = bndtobnd(psi, U, dpar[f], dws, lp)
        flush(log_file)

        println(log_file,"Contracting propagators...")
        flush(log_file)

        f1_sf[f] = ctoreal(norm2(C11) + norm2(C21) + norm2(C31) + norm2(C12) + norm2(C22) + norm2(C32))

        k1_sf[f] = (dot(C11,dmul(Gamma{6},imm(C12))) + dot(C11,dmul(Gamma{7},-C12)) + dot(C11,dmul(Gamma{8}, imm(C11)))
                    +dot(C12,dmul(Gamma{6},imm(C11))) + dot(C12,dmul(Gamma{7}, C11)) + dot(C12,dmul(Gamma{8},mimm(C12)))
                    +dot(C21,dmul(Gamma{6},imm(C22))) + dot(C21,dmul(Gamma{7},-C22)) + dot(C21,dmul(Gamma{8}, imm(C21)))
                    +dot(C22,dmul(Gamma{6},imm(C21))) + dot(C22,dmul(Gamma{7}, C21)) + dot(C22,dmul(Gamma{8},mimm(C22)))
                    +dot(C31,dmul(Gamma{6},imm(C32))) + dot(C31,dmul(Gamma{7},-C32)) + dot(C31,dmul(Gamma{8}, imm(C31)))
                    +dot(C32,dmul(Gamma{6},imm(C31))) + dot(C32,dmul(Gamma{7}, C31)) + dot(C32,dmul(Gamma{8},mimm(C32))) )/3

        fp_sf[:,f] .= zero(Series{Float64,Nder})
        fa_sf[:,f] .= zero(Series{Complex{Float64},Nder})
        kt_sf[:,f] .= zero(Series{Complex{Float64},Nder})
        kv_sf[:,f] .= zero(Series{Complex{Float64},Nder})

        @timeit "Volume sum" for b in 1:lp.bsz for r in 1:lp.rsz
            t = point_time((b,r),lp)

            fp_sf[t,f] += ctoreal(norm2(A11[b,r]) + norm2(A21[b,r]) + norm2(A31[b,r]) + norm2(A12[b,r]) + norm2(A22[b,r]) + norm2(A32[b,r]))/prod(lp.iL[1:3])
            fa_sf[t,f] -= (dot(A11[b,r],dmul(Gamma{4},A11[b,r])) + dot(A21[b,r],dmul(Gamma{4},A21[b,r])) + dot(A31[b,r],dmul(Gamma{4},A31[b,r]))
                           + dot(A12[b,r],dmul(Gamma{4},A12[b,r])) + dot(A22[b,r],dmul(Gamma{4},A22[b,r])) + dot(A32[b,r],dmul(Gamma{4},A32[b,r])))/prod(lp.iL[1:3])
            kv_sf[t,f] += ( dot(A11[b,r],dmul(Gamma{6},imm(A12[b,r]))) + dot(A11[b,r],dmul(Gamma{7},-A12[b,r])) + dot(A11[b,r],dmul(Gamma{8}, imm(A11[b,r])))
                            +dot(A12[b,r],dmul(Gamma{6},imm(A11[b,r]))) + dot(A12[b,r],dmul(Gamma{7}, A11[b,r])) + dot(A12[b,r],dmul(Gamma{8},mimm(A12[b,r])))
                            +dot(A21[b,r],dmul(Gamma{6},imm(A22[b,r]))) + dot(A21[b,r],dmul(Gamma{7},-A22[b,r])) + dot(A21[b,r],dmul(Gamma{8}, imm(A21[b,r])))
                            +dot(A22[b,r],dmul(Gamma{6},imm(A21[b,r]))) + dot(A22[b,r],dmul(Gamma{7}, A21[b,r])) + dot(A22[b,r],dmul(Gamma{8},mimm(A22[b,r])))
                            +dot(A31[b,r],dmul(Gamma{6},imm(A32[b,r]))) + dot(A31[b,r],dmul(Gamma{7},-A32[b,r])) + dot(A31[b,r],dmul(Gamma{8}, imm(A31[b,r])))
                            +dot(A32[b,r],dmul(Gamma{6},imm(A31[b,r]))) + dot(A32[b,r],dmul(Gamma{7}, A31[b,r])) + dot(A32[b,r],dmul(Gamma{8},mimm(A32[b,r]))) )/(3*prod(lp.iL[1:3]))
            kt_sf[t,f] +=im*(dot(A11[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A12[b,r])))) + dot(A11[b,r],dmul(Gamma{11},dmul(Gamma{5},-A12[b,r]))) + dot(A11[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A11[b,r]))))
                             +dot(A12[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A11[b,r])))) + dot(A12[b,r],dmul(Gamma{11},dmul(Gamma{5}, A11[b,r]))) + dot(A12[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A12[b,r]))))
                             +dot(A21[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A22[b,r])))) + dot(A21[b,r],dmul(Gamma{11},dmul(Gamma{5},-A22[b,r]))) + dot(A21[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A21[b,r]))))
                             +dot(A22[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A21[b,r])))) + dot(A22[b,r],dmul(Gamma{11},dmul(Gamma{5}, A21[b,r]))) + dot(A22[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A22[b,r]))))
                             +dot(A31[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A32[b,r])))) + dot(A31[b,r],dmul(Gamma{11},dmul(Gamma{5},-A32[b,r]))) + dot(A31[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A31[b,r]))))
                             +dot(A32[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A31[b,r])))) + dot(A32[b,r],dmul(Gamma{11},dmul(Gamma{5}, A31[b,r]))) + dot(A32[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A32[b,r])))) )/(3*prod(lp.iL[1:3]))
        end end

        # println(log_file,"Computing propagators from t=T to the bulk")
        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 1, 1)
        # println(log_file,"CG converged for c=1 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B11 .= Array(psi)
        # flush(log_file)

        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 2, 1)
        # println(log_file,"CG converged for c=2 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B21 .= Array(psi)
        # flush(log_file)

        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 3, 1)
        # println(log_file,"CG converged for c=3 and s=1 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B31 .= Array(psi)
        # flush(log_file)

        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 1, 2)
        # println(log_file,"CG converged for c=1 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B12 .= Array(psi)
        # flush(log_file)

        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 2, 2)
        # println(log_file,"CG converged for c=2 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B22 .= Array(psi)
        # flush(log_file)

        # niter = Tbndpropagator!(psi, U, dpar, dws, lp, params["Fermion"]["maxiter"], params["Fermion"]["tolerance"], 3, 2)
        # println(log_file,"CG converged for c=3 and s=2 after ",niter," iterations with absolute residue ", CUDA.mapreduce(x -> norm(x), +, dws.sr))
        # global B32 .= Array(psi)
        # flush(log_file)


        # println(log_file,"\nComputing gP...\n")
        # flush(log_file)
        # global gP = gP_fun()

        # println(log_file,"\nComputing gA...\n")
        # flush(log_file)
        # global gA = gA_fun()


        # println(log_file,"\nComputing lV...\n")
        # flush(log_file)
        # global lV = lV_fun()

        # println(log_file,"\nComputing lT...\n")
        # flush(log_file)
        # global lT = lT_fun()

    end

    return nothing
end


# function gP_fun()
#     cf = Vector{Complex{Float64}}(undef,lp.iL[4])
#     for t in 1:lp.iL[4]
#         cf[t] = 0.0
#         for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
#             b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
#             cf[t] += (norm2(B11[b,r]) + norm2(B21[b,r]) + norm2(B31[b,r]) + norm2(B12[b,r]) + norm2(B22[b,r]) + norm2(B32[b,r]))/prod(lp.iL[1:3])
#         end end end
#     end
#     return cf
# end

# function gA_fun()
#     cf = Vector{Complex{Float64}}(undef,lp.iL[4])
#     for t in 1:lp.iL[4]
#         cf[t] = 0.0
#         for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
#             b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
#             cf[t] += (dot(B11[b,r],dmul(Gamma{4},B11[b,r])) + dot(B21[b,r],dmul(Gamma{4},B21[b,r])) + dot(B31[b,r],dmul(Gamma{4},B31[b,r]))
#                       + dot(B12[b,r],dmul(Gamma{4},B12[b,r])) + dot(B22[b,r],dmul(Gamma{4},B22[b,r])) + dot(B32[b,r],dmul(Gamma{4},B32[b,r])))/prod(lp.iL[1:3])
#         end end end
#     end
#     return cf
# end


# function lV_fun()
#     cf = Vector{Complex{Float64}}(undef,lp.iL[4])
#     for t in 1:lp.iL[4]
#         cf[t] = 0.0
#         for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
#             b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
#             cf[t] += ( dot(B11[b,r],dmul(Gamma{6},mimm(B12[b,r]))) + dot(B11[b,r],dmul(Gamma{7}, B12[b,r])) + dot(B11[b,r],dmul(Gamma{8},mimm(B11[b,r])))
#                        +dot(B12[b,r],dmul(Gamma{6},mimm(B11[b,r]))) + dot(B12[b,r],dmul(Gamma{7},-B11[b,r])) + dot(B12[b,r],dmul(Gamma{8}, imm(B12[b,r])))
#                        +dot(B21[b,r],dmul(Gamma{6},mimm(B22[b,r]))) + dot(B21[b,r],dmul(Gamma{7}, B22[b,r])) + dot(B21[b,r],dmul(Gamma{8},mimm(B21[b,r])))
#                        +dot(B22[b,r],dmul(Gamma{6},mimm(B21[b,r]))) + dot(B22[b,r],dmul(Gamma{7},-B21[b,r])) + dot(B22[b,r],dmul(Gamma{8}, imm(B22[b,r])))
#                        +dot(B31[b,r],dmul(Gamma{6},mimm(B32[b,r]))) + dot(B31[b,r],dmul(Gamma{7}, B32[b,r])) + dot(B31[b,r],dmul(Gamma{8},mimm(B31[b,r])))
#                        +dot(B32[b,r],dmul(Gamma{6},mimm(B31[b,r]))) + dot(B32[b,r],dmul(Gamma{7},-B31[b,r])) + dot(B32[b,r],dmul(Gamma{8}, imm(B32[b,r]))) )/(3*prod(lp.iL[1:3]))
#         end end end
#     end
#     return cf
# end

# function lT_fun()
#     cf = Vector{Complex{Float64}}(undef,lp.iL[4])
#     for t in 1:lp.iL[4]
#         cf[t] = 0.0
#         for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
#             b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
#             cf[t] +=-im*(dot(B11[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B12[b,r])))) + dot(B11[b,r],dmul(Gamma{11},dmul(Gamma{5}, B12[b,r]))) + dot(B11[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B11[b,r]))))
#                          +dot(B12[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B11[b,r])))) + dot(B12[b,r],dmul(Gamma{11},dmul(Gamma{5},-B11[b,r]))) + dot(B12[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B12[b,r]))))
#                          +dot(B21[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B22[b,r])))) + dot(B21[b,r],dmul(Gamma{11},dmul(Gamma{5}, B22[b,r]))) + dot(B21[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B21[b,r]))))
#                          +dot(B22[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B21[b,r])))) + dot(B22[b,r],dmul(Gamma{11},dmul(Gamma{5},-B21[b,r]))) + dot(B22[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B22[b,r]))))
#                          +dot(B31[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B32[b,r])))) + dot(B31[b,r],dmul(Gamma{11},dmul(Gamma{5}, B32[b,r]))) + dot(B31[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B31[b,r]))))
#                          +dot(B32[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B31[b,r])))) + dot(B32[b,r],dmul(Gamma{11},dmul(Gamma{5},-B31[b,r]))) + dot(B32[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B32[b,r])))) )/(3*prod(lp.iL[1:3]))
#         end end end
#     end
#     return cf
# end
